@@ -1,10 +1,9 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2006
- *	Oracle Corporation.  All rights reserved.
+ * Copyright (c) 1996, 2010 Oracle and/or its affiliates.  All rights reserved.
  *
- * $Id: env_recover.c,v 12.34 2006/09/09 14:28:22 bostic Exp $
+ * $Id$
  */
 
 #include "db_config.h"
@@ -14,22 +13,26 @@
 #include "dbinc/fop.h"
 #include "dbinc/btree.h"
 #include "dbinc/hash.h"
-#include "dbinc/log.h"
 #include "dbinc/mp.h"
 #include "dbinc/qam.h"
 #include "dbinc/txn.h"
 
 #ifndef lint
 static const char copyright[] =
-    "Copyright (c) 1996-2006\nOracle Corporation.  All rights reserved.\n";
+    "Copyright (c) 1996, 2010 Oracle and/or its affiliates.  All rights reserved.\n";
 #endif
 
-static int	__db_log_corrupt __P((DB_ENV *, DB_LSN *));
-static int	__env_init_rec_42 __P((DB_ENV *));
-static int	__env_init_rec_43 __P((DB_ENV *));
-static int	__env_init_rec_45 __P((DB_ENV *));
-static int	__log_earliest __P((DB_ENV *, DB_LOGC *, int32_t *, DB_LSN *));
+static int	__db_log_corrupt __P((ENV *, DB_LSN *));
+static int	__env_init_rec_42 __P((ENV *));
+static int	__env_init_rec_43 __P((ENV *));
+static int	__env_init_rec_46 __P((ENV *));
+static int	__env_init_rec_47 __P((ENV *));
+static int	__env_init_rec_48 __P((ENV *));
+static int	__log_earliest __P((ENV *, DB_LOGC *, int32_t *, DB_LSN *));
+
+#ifndef HAVE_BREW
 static double	__lsn_diff __P((DB_LSN *, DB_LSN *, DB_LSN *, u_int32_t, int));
+#endif
 
 /*
  * __db_apprec --
@@ -38,18 +41,22 @@ static double	__lsn_diff __P((DB_LSN *, DB_LSN *, DB_LSN *, u_int32_t, int));
  * LSN of max_lsn, so we need to roll back sufficiently far for that
  * to work.  See __log_backup for details.
  *
- * PUBLIC: int __db_apprec __P((DB_ENV *, DB_LSN *, DB_LSN *, int, u_int32_t));
+ * PUBLIC: int __db_apprec __P((ENV *,
+ * PUBLIC:     DB_THREAD_INFO *, DB_LSN *, DB_LSN *, int, u_int32_t));
  */
 int
-__db_apprec(dbenv, max_lsn, trunclsn, update, flags)
-	DB_ENV *dbenv;
+__db_apprec(env, ip, max_lsn, trunclsn, update, flags)
+	ENV *env;
+	DB_THREAD_INFO *ip;
 	DB_LSN *max_lsn, *trunclsn;
 	int update;
 	u_int32_t flags;
 {
 	DBT data;
+	DB_ENV *dbenv;
 	DB_LOGC *logc;
 	DB_LSN ckp_lsn, first_lsn, last_lsn, lowlsn, lsn, stop_lsn, tlsn;
+	DB_LSN *vtrunc_ckp, *vtrunc_lsn;
 	DB_TXNHEAD *txninfo;
 	DB_TXNREGION *region;
 	REGENV *renv;
@@ -59,12 +66,13 @@ __db_apprec(dbenv, max_lsn, trunclsn, update, flags)
 	double nfiles;
 	u_int32_t hi_txn, log_size, txnid;
 	int32_t low;
-	int have_rec, progress, ret, t_ret;
+	int all_recovered, have_rec, progress, ret, t_ret;
 	char *p, *pass;
 	char t1[CTIME_BUFLEN], t2[CTIME_BUFLEN], time_buf[CTIME_BUFLEN];
 
 	COMPQUIET(nfiles, (double)0.001);
 
+	dbenv = env->dbenv;
 	logc = NULL;
 	ckp_args = NULL;
 	hi_txn = TXN_MAXIMUM;
@@ -77,24 +85,24 @@ __db_apprec(dbenv, max_lsn, trunclsn, update, flags)
 	 * Get the log size.  No locking required because we're single-threaded
 	 * during recovery.
 	 */
-	log_size = ((LOG *)dbenv->lg_handle->reginfo.primary)->log_size;
+	log_size = ((LOG *)env->lg_handle->reginfo.primary)->log_size;
 
 	/*
 	 * If we need to, update the env handle timestamp.
 	 */
-	if (update && REP_ON(dbenv)) {
-		infop = dbenv->reginfo;
+	if (update && REP_ON(env)) {
+		infop = env->reginfo;
 		renv = infop->primary;
 		(void)time(&renv->rep_timestamp);
 	}
 
 	/* Set in-recovery flags. */
-	F_SET(dbenv->lg_handle, DBLOG_RECOVER);
-	region = dbenv->tx_handle->reginfo.primary;
+	F_SET(env->lg_handle, DBLOG_RECOVER);
+	region = env->tx_handle->reginfo.primary;
 	F_SET(region, TXN_IN_RECOVERY);
 
 	/* Allocate a cursor for the log. */
-	if ((ret = __log_cursor(dbenv, &logc)) != 0)
+	if ((ret = __log_cursor(env, &logc)) != 0)
 		goto err;
 
 	/*
@@ -103,27 +111,27 @@ __db_apprec(dbenv, max_lsn, trunclsn, update, flags)
 	 */
 	ZERO_LSN(lowlsn);
 	if (max_lsn != NULL) {
-		if ((ret = __log_backup(dbenv, logc, max_lsn, &lowlsn,
+		if ((ret = __log_backup(env, logc, max_lsn, &lowlsn,
 		    CKPLSN_CMP)) != 0)
 			goto err;
 	} else if (dbenv->tx_timestamp != 0) {
-		if ((ret = __log_earliest(dbenv, logc, &low, &lowlsn)) != 0)
+		if ((ret = __log_earliest(env, logc, &low, &lowlsn)) != 0)
 			goto err;
 		if ((int32_t)dbenv->tx_timestamp < low) {
 			t1[sizeof(t1) - 1] = '\0';
-			(void)strncpy(t1, __db_ctime(
+			(void)strncpy(t1, __os_ctime(
 			    &dbenv->tx_timestamp, time_buf), sizeof(t1) - 1);
 			if ((p = strchr(t1, '\n')) != NULL)
 				*p = '\0';
 
 			t2[sizeof(t2) - 1] = '\0';
 			tlow = (time_t)low;
-			(void)strncpy(t2, __db_ctime(
+			(void)strncpy(t2, __os_ctime(
 			    &tlow, time_buf), sizeof(t2) - 1);
 			if ((p = strchr(t2, '\n')) != NULL)
 				*p = '\0';
 
-			__db_errx(dbenv,
+			__db_errx(env,
 		    "Invalid recovery timestamp %s; earliest time is %s",
 			    t1, t2);
 			ret = EINVAL;
@@ -204,22 +212,22 @@ __db_apprec(dbenv, max_lsn, trunclsn, update, flags)
 	ZERO_LSN(last_lsn);
 #endif
 	memset(&data, 0, sizeof(data));
-	if ((ret = __log_c_get(logc, &last_lsn, &data, DB_LAST)) != 0) {
+	if ((ret = __logc_get(logc, &last_lsn, &data, DB_LAST)) != 0) {
 		if (ret == DB_NOTFOUND)
 			ret = 0;
 		else
-			__db_errx(dbenv, "Last log record not found");
+			__db_errx(env, "Last log record not found");
 		goto err;
 	}
 
 	do {
 		/* txnid is after rectype, which is a u_int32. */
-		memcpy(&txnid,
-		    (u_int8_t *)data.data + sizeof(u_int32_t), sizeof(txnid));
+		LOGCOPY_32(env, &txnid,
+		    (u_int8_t *)data.data + sizeof(u_int32_t));
 
 		if (txnid != 0)
 			break;
-	} while ((ret = __log_c_get(logc, &lsn, &data, DB_PREV)) == 0);
+	} while ((ret = __logc_get(logc, &lsn, &data, DB_PREV)) == 0);
 
 	/*
 	 * There are no transactions, so there is nothing to do unless
@@ -249,30 +257,30 @@ __db_apprec(dbenv, max_lsn, trunclsn, update, flags)
 	 * Get the first LSN in the log; it's an initial default
 	 * even if this is not a catastrophic recovery.
 	 */
-	if ((ret = __log_c_get(logc, &ckp_lsn, &data, DB_FIRST)) != 0) {
+	if ((ret = __logc_get(logc, &ckp_lsn, &data, DB_FIRST)) != 0) {
 		if (ret == DB_NOTFOUND)
 			ret = 0;
 		else
-			__db_errx(dbenv, "First log record not found");
+			__db_errx(env, "First log record not found");
 		goto err;
 	}
 	first_lsn = ckp_lsn;
 	have_rec = 1;
 
 	if (!LF_ISSET(DB_RECOVER_FATAL)) {
-		if ((ret = __txn_getckp(dbenv, &ckp_lsn)) == 0 &&
-		    (ret = __log_c_get(logc, &ckp_lsn, &data, DB_SET)) == 0) {
+		if ((ret = __txn_getckp(env, &ckp_lsn)) == 0 &&
+		    (ret = __logc_get(logc, &ckp_lsn, &data, DB_SET)) == 0) {
 			/* We have a recent checkpoint.  This is LSN (1). */
-			if ((ret = __txn_ckp_read(dbenv,
+			if ((ret = __txn_ckp_read(env,
 			    data.data, &ckp_args)) != 0) {
-				__db_errx(dbenv,
+				__db_errx(env,
 			    "Invalid checkpoint record at [%ld][%ld]",
 				    (u_long)ckp_lsn.file,
 				    (u_long)ckp_lsn.offset);
 				goto err;
 			}
 			first_lsn = ckp_args->ckp_lsn;
-			__os_free(dbenv, ckp_args);
+			__os_free(env, ckp_args);
 			have_rec = 0;
 		}
 
@@ -296,19 +304,20 @@ __db_apprec(dbenv, max_lsn, trunclsn, update, flags)
 		 */
 		if ((dbenv->tx_timestamp != 0 || max_lsn != NULL) &&
 		    LOG_COMPARE(&lowlsn, &first_lsn) < 0) {
-			DB_ASSERT(dbenv, have_rec == 0);
+			DB_ASSERT(env, have_rec == 0);
 			first_lsn = lowlsn;
 		}
 	}
 
 	/* Get the record at first_lsn if we don't have it already. */
 	if (!have_rec &&
-	    (ret = __log_c_get(logc, &first_lsn, &data, DB_SET)) != 0) {
-		__db_errx(dbenv, "Checkpoint LSN record [%ld][%ld] not found",
+	    (ret = __logc_get(logc, &first_lsn, &data, DB_SET)) != 0) {
+		__db_errx(env, "Checkpoint LSN record [%ld][%ld] not found",
 		    (u_long)first_lsn.file, (u_long)first_lsn.offset);
 		goto err;
 	}
 
+#ifndef HAVE_BREW
 	if (dbenv->db_feedback != NULL) {
 		if (last_lsn.file == first_lsn.file)
 			nfiles = (double)
@@ -321,17 +330,18 @@ __db_apprec(dbenv, max_lsn, trunclsn, update, flags)
 		if (nfiles < 0.001)
 			nfiles = 0.001;
 	}
+#endif
 
 	/* Find a low txnid. */
 	ret = 0;
 	if (hi_txn != 0) do {
 		/* txnid is after rectype, which is a u_int32. */
-		memcpy(&txnid,
-		    (u_int8_t *)data.data + sizeof(u_int32_t), sizeof(txnid));
+		LOGCOPY_32(env, &txnid,
+		    (u_int8_t *)data.data + sizeof(u_int32_t));
 
 		if (txnid != 0)
 			break;
-	} while ((ret = __log_c_get(logc, &lsn, &data, DB_NEXT)) == 0);
+	} while ((ret = __logc_get(logc, &lsn, &data, DB_NEXT)) == 0);
 
 	/*
 	 * There are no transactions and we're not recovering to an LSN (see
@@ -339,32 +349,34 @@ __db_apprec(dbenv, max_lsn, trunclsn, update, flags)
 	 */
 	if (ret == DB_NOTFOUND) {
 		if (LOG_COMPARE(&lsn, &last_lsn) != 0)
-			ret = __db_log_corrupt(dbenv, &lsn);
+			ret = __db_log_corrupt(env, &lsn);
 		else
 			ret = 0;
 	}
 
 	/* Reset to the first lsn. */
 	if (ret != 0 ||
-	    (ret = __log_c_get(logc, &first_lsn, &data, DB_SET)) != 0)
+	    (ret = __logc_get(logc, &first_lsn, &data, DB_SET)) != 0)
 		goto err;
 
 	/* Initialize the transaction list. */
-	if ((ret =
-	    __db_txnlist_init(dbenv, txnid, hi_txn, max_lsn, &txninfo)) != 0)
+	if ((ret = __db_txnlist_init(env, ip,
+	    txnid, hi_txn, max_lsn, &txninfo)) != 0)
 		goto err;
 
 	/*
 	 * Pass #1
 	 * Run forward through the log starting at the first relevant lsn.
 	 */
-	if ((ret = __env_openfiles(dbenv, logc,
+	if ((ret = __env_openfiles(env, logc,
 	    txninfo, &data, &first_lsn, &last_lsn, nfiles, 1)) != 0)
 		goto err;
 
 	/* If there were no transactions, then we can bail out early. */
-	if (hi_txn == 0 && max_lsn == NULL)
+	if (hi_txn == 0 && max_lsn == NULL) {
+		lsn = last_lsn;
 		goto done;
+	}
 
 	/*
 	 * Pass #2.
@@ -372,24 +384,26 @@ __db_apprec(dbenv, max_lsn, trunclsn, update, flags)
 	 * We used first_lsn to tell us how far back we need to recover,
 	 * use it here.
 	 */
-
 	if (FLD_ISSET(dbenv->verbose, DB_VERB_RECOVERY))
-		__db_msg(dbenv, "Recovery starting from [%lu][%lu]",
+		__db_msg(env, "Recovery starting from [%lu][%lu]",
 		    (u_long)first_lsn.file, (u_long)first_lsn.offset);
 
 	pass = "backward";
-	for (ret = __log_c_get(logc, &lsn, &data, DB_LAST);
+	for (ret = __logc_get(logc, &lsn, &data, DB_LAST);
 	    ret == 0 && LOG_COMPARE(&lsn, &first_lsn) >= 0;
-	    ret = __log_c_get(logc, &lsn, &data, DB_PREV)) {
+	    ret = __logc_get(logc, &lsn, &data, DB_PREV)) {
+#ifdef HAVE_BREW
+		COMPQUIET(progress, 0);
+#else
 		if (dbenv->db_feedback != NULL) {
 			progress = 34 + (int)(33 * (__lsn_diff(&first_lsn,
 			    &last_lsn, &lsn, log_size, 0) / nfiles));
 			dbenv->db_feedback(dbenv, DB_RECOVER, progress);
 		}
+#endif
 		tlsn = lsn;
-		ret = __db_dispatch(dbenv, dbenv->recover_dtab,
-		    dbenv->recover_dtab_size, &data, &tlsn,
-		    DB_TXN_BACKWARD_ROLL, txninfo);
+		ret = __db_dispatch(env, &env->recover_dtab,
+		    &data, &tlsn, DB_TXN_BACKWARD_ROLL, txninfo);
 		if (ret != 0) {
 			if (ret != DB_TXN_CKP)
 				goto msgerr;
@@ -399,7 +413,7 @@ __db_apprec(dbenv, max_lsn, trunclsn, update, flags)
 	}
 	if (ret == DB_NOTFOUND) {
 		if (LOG_COMPARE(&lsn, &first_lsn) > 0)
-			ret = __db_log_corrupt(dbenv, &lsn);
+			ret = __db_log_corrupt(env, &lsn);
 		else
 			ret = 0;
 	}
@@ -421,17 +435,18 @@ __db_apprec(dbenv, max_lsn, trunclsn, update, flags)
 	if (max_lsn != NULL || dbenv->tx_timestamp != 0)
 		stop_lsn = ((DB_TXNHEAD *)txninfo)->maxlsn;
 
-	for (ret = __log_c_get(logc, &lsn, &data, DB_NEXT);
-	    ret == 0; ret = __log_c_get(logc, &lsn, &data, DB_NEXT)) {
+	for (ret = __logc_get(logc, &lsn, &data, DB_NEXT);
+	    ret == 0; ret = __logc_get(logc, &lsn, &data, DB_NEXT)) {
+#ifndef HAVE_BREW
 		if (dbenv->db_feedback != NULL) {
 			progress = 67 + (int)(33 * (__lsn_diff(&first_lsn,
 			    &last_lsn, &lsn, log_size, 1) / nfiles));
 			dbenv->db_feedback(dbenv, DB_RECOVER, progress);
 		}
+#endif
 		tlsn = lsn;
-		ret = __db_dispatch(dbenv, dbenv->recover_dtab,
-		    dbenv->recover_dtab_size, &data, &tlsn,
-		    DB_TXN_FORWARD_ROLL, txninfo);
+		ret = __db_dispatch(env, &env->recover_dtab,
+		    &data, &tlsn, DB_TXN_FORWARD_ROLL, txninfo);
 		if (ret != 0) {
 			if (ret != DB_TXN_CKP)
 				goto msgerr;
@@ -448,144 +463,154 @@ __db_apprec(dbenv, max_lsn, trunclsn, update, flags)
 
 	}
 	if (ret == DB_NOTFOUND)
-		ret = __db_log_corrupt(dbenv, &lsn);
+		ret = __db_log_corrupt(env, &lsn);
 	if (ret != 0)
 		goto err;
 
-#ifndef HAVE_FTRUNCATE
-	/*
-	 * Process any pages that were on the limbo list and move them to
-	 * the free list.  Do this before checkpointing the database.
-	 */
-	if ((ret = __db_do_the_limbo(dbenv, NULL, NULL, txninfo,
-	      dbenv->tx_timestamp != 0 ? LIMBO_TIMESTAMP : LIMBO_RECOVER)) != 0)
-		goto err;
-#endif
-
 	if (max_lsn == NULL)
 		region->last_txnid = ((DB_TXNHEAD *)txninfo)->maxid;
 
-	if (dbenv->tx_timestamp != 0) {
-		/* We are going to truncate, so we'd best close the cursor. */
-		if (logc != NULL && (ret = __log_c_close(logc)) != 0)
+done:
+	/* We are going to truncate, so we'd best close the cursor. */
+	if (logc != NULL) {
+		if ((ret = __logc_close(logc)) != 0)
 			goto err;
 		logc = NULL;
-		/* Flush everything to disk, we are losing the log. */
-		if ((ret = __memp_sync(dbenv, NULL)) != 0)
-			goto err;
-		region->last_ckp = ((DB_TXNHEAD *)txninfo)->ckplsn;
-		if ((ret = __log_vtruncate(dbenv,
-		    &((DB_TXNHEAD *)txninfo)->maxlsn,
-		    &((DB_TXNHEAD *)txninfo)->ckplsn, trunclsn)) != 0)
-			goto err;
-
-#ifndef HAVE_FTRUNCATE
-		/*
-		 * Generate logging compensation records.
-		 * If we crash during/after vtruncate we may have
-		 * pages missing from the free list since they
-		 * if we roll things further back from here.
-		 * These pages are only known in memory at this pont.
-		 */
-		 if ((ret = __db_do_the_limbo(dbenv,
-		       NULL, NULL, txninfo, LIMBO_COMPENSATE)) != 0)
-			goto err;
-#endif
 	}
-
-done:
-	/* Take a checkpoint here to force any dirty data pages to disk. */
-	if ((ret = __txn_checkpoint(dbenv, 0, 0, DB_FORCE)) != 0) {
+	/*
+	 * Also flush the cache before truncating the log. It's recovery,
+	 * ignore any application max-write configuration.
+	 */
+	if ((ret = __memp_sync_int(env,
+	    NULL, 0, DB_SYNC_CACHE | DB_SYNC_SUPPRESS_WRITE, NULL, NULL)) != 0)
+		goto err;
+	if (dbenv->tx_timestamp != 0) {
+		/* Run recovery up to this timestamp. */
+		region->last_ckp = ((DB_TXNHEAD *)txninfo)->ckplsn;
+		vtrunc_lsn = &((DB_TXNHEAD *)txninfo)->maxlsn;
+		vtrunc_ckp = &((DB_TXNHEAD *)txninfo)->ckplsn;
+	} else if (max_lsn != NULL) {
 		/*
-		 * If there was no space for the checkpoint we can
-		 * still bring the environment up.  No updates will
-		 * be able to commit either, but the environment can
-		 * be used read only.
+		 * Flush everything to disk, we are losing the log.  It's
+		 * recovery, ignore any application max-write configuration.
 		 */
-		if (max_lsn == NULL && ret == ENOSPC)
-			ret = 0;
+		if ((ret = __memp_sync_int(env, NULL, 0,
+		    DB_SYNC_CACHE | DB_SYNC_SUPPRESS_WRITE, NULL, NULL)) != 0)
+			goto err;
+		/* This is a HA client syncing to the master. */
+		if (!IS_ZERO_LSN(((DB_TXNHEAD *)txninfo)->ckplsn))
+			region->last_ckp = ((DB_TXNHEAD *)txninfo)->ckplsn;
+		else if ((ret =
+		    __txn_findlastckp(env, &region->last_ckp, max_lsn)) != 0)
+			goto err;
+		vtrunc_lsn = max_lsn;
+		vtrunc_ckp = &((DB_TXNHEAD *)txninfo)->ckplsn;
+	} else {
+		/*
+		 * The usual case: we recovered the whole (valid) log; clear
+		 * out any partial record after the recovery point.
+		 */
+		vtrunc_lsn = &lsn;
+		vtrunc_ckp = &region->last_ckp;
+	}
+	if ((ret = __log_vtruncate(env, vtrunc_lsn, vtrunc_ckp, trunclsn)) != 0)
+		goto err;
+
+	/*
+	 * Usually we close all files at the end of recovery, unless there are
+	 * prepared transactions or errors in the checkpoint.
+	 */
+	all_recovered = region->stat.st_nrestores == 0;
+	/*
+	 * Log a checkpoint here so subsequent recoveries can skip what's been
+	 * done; this is unnecessary for HA rep clients, as they do not write
+	 * log records.
+	 */
+	if (max_lsn == NULL && (ret = __txn_checkpoint(env,
+	    0, 0, DB_CKP_INTERNAL | DB_FORCE)) != 0) {
+		/*
+		 * If there was no space for the checkpoint or flushng db
+		 * pages we can still bring the environment up, if only for
+		 * read-only access. We must not close the open files because a
+		 * subsequent recovery might still need to redo this portion
+		 * of the log [#18590].
+		 */
+		if (max_lsn == NULL && ret == ENOSPC) {
+			if (FLD_ISSET(dbenv->verbose, DB_VERB_RECOVERY))
+				__db_msg(env,
+		    "Recovery continuing after non-fatal checkpoint error: %s",
+				    db_strerror(ret));
+			all_recovered = 0;
+		}
 		else
 			goto err;
 	}
 
-	/* Close all the db files that are open. */
-	if ((ret = __dbreg_close_files(dbenv)) != 0)
-		goto err;
+	if (all_recovered ) {
+		/* Close all the db files that are open. */
+		if ((ret = __dbreg_close_files(env, 0)) != 0)
+			goto err;
+	} else {
+		if ((ret = __dbreg_mark_restored(env)) != 0)
+			goto err;
+		F_SET(env->lg_handle, DBLOG_OPENFILES);
+	}
 
 	if (max_lsn != NULL) {
-		if (!IS_ZERO_LSN(((DB_TXNHEAD *)txninfo)->ckplsn))
-			region->last_ckp = ((DB_TXNHEAD *)txninfo)->ckplsn;
-		else if ((ret =
-		    __txn_findlastckp(dbenv, &region->last_ckp, max_lsn)) != 0)
-			goto err;
-
-		/* We are going to truncate, so we'd best close the cursor. */
-		if (logc != NULL && (ret = __log_c_close(logc)) != 0)
-			goto err;
-		logc = NULL;
-		if ((ret = __log_vtruncate(dbenv,
-		    max_lsn, &((DB_TXNHEAD *)txninfo)->ckplsn, trunclsn)) != 0)
-			goto err;
-
 		/*
 		 * Now we need to open files that should be open in order for
 		 * client processing to continue.  However, since we've
 		 * truncated the log, we need to recompute from where the
 		 * openfiles pass should begin.
 		 */
-		if ((ret = __log_cursor(dbenv, &logc)) != 0)
+		if ((ret = __log_cursor(env, &logc)) != 0)
 			goto err;
 		if ((ret =
-		    __log_c_get(logc, &first_lsn, &data, DB_FIRST)) != 0) {
+		    __logc_get(logc, &first_lsn, &data, DB_FIRST)) != 0) {
 			if (ret == DB_NOTFOUND)
 				ret = 0;
 			else
-				__db_errx(dbenv, "First log record not found");
+				__db_errx(env, "First log record not found");
 			goto err;
 		}
-		if ((ret = __txn_getckp(dbenv, &first_lsn)) == 0 &&
-		    (ret = __log_c_get(logc, &first_lsn, &data, DB_SET)) == 0) {
+		if ((ret = __txn_getckp(env, &first_lsn)) == 0 &&
+		    (ret = __logc_get(logc, &first_lsn, &data, DB_SET)) == 0) {
 			/* We have a recent checkpoint.  This is LSN (1). */
-			if ((ret = __txn_ckp_read(dbenv,
+			if ((ret = __txn_ckp_read(env,
 			    data.data, &ckp_args)) != 0) {
-				__db_errx(dbenv,
+				__db_errx(env,
 			    "Invalid checkpoint record at [%ld][%ld]",
 				    (u_long)first_lsn.file,
 				    (u_long)first_lsn.offset);
 				goto err;
 			}
 			first_lsn = ckp_args->ckp_lsn;
-			__os_free(dbenv, ckp_args);
+			__os_free(env, ckp_args);
 		}
-		if ((ret = __log_c_get(logc, &first_lsn, &data, DB_SET)) != 0)
+		if ((ret = __logc_get(logc, &first_lsn, &data, DB_SET)) != 0)
 			goto err;
-		if ((ret = __env_openfiles(dbenv, logc,
+		if ((ret = __env_openfiles(env, logc,
 		    txninfo, &data, &first_lsn, max_lsn, nfiles, 1)) != 0)
 			goto err;
-	} else if (region->stat.st_nrestores == 0) {
+	} else if (all_recovered) {
 		/*
-		 * If there are no prepared transactions that need resolution,
-		 * we need to reset the transaction ID space and log this fact.
+		 * If there are no transactions that need resolution, whether
+		 * because they are prepared or because recovery will need to
+		 * process them, we need to reset the transaction ID space and
+		 * log this fact.
 		 */
-		if ((ret = __txn_reset(dbenv)) != 0)
+		if ((ret = __txn_reset(env)) != 0)
 			goto err;
 	} else {
-		/*
-		 * If we have restored prepared txns then they are in process
-		 * as far as replication is concerned.
-		 */
-		if (REP_ON(dbenv))
-			dbenv->rep_handle->region->op_cnt =
-			    region->stat.st_nrestores;
-		if ((ret = __txn_recycle_id(dbenv)) != 0)
+		if ((ret = __txn_recycle_id(env)) != 0)
 			goto err;
 	}
 
 	if (FLD_ISSET(dbenv->verbose, DB_VERB_RECOVERY)) {
 		(void)time(&now);
-		__db_msg(dbenv,
-		    "Recovery complete at %.24s", __db_ctime(&now, time_buf));
-		__db_msg(dbenv, "%s %lx %s [%lu][%lu]",
+		__db_msg(env,
+		    "Recovery complete at %.24s", __os_ctime(&now, time_buf));
+		__db_msg(env, "%s %lx %s [%lu][%lu]",
 		    "Maximum transaction ID",
 		    (u_long)(txninfo == NULL ?
 			TXN_MINIMUM : ((DB_TXNHEAD *)txninfo)->maxid),
@@ -595,25 +620,26 @@ done:
 	}
 
 	if (0) {
-msgerr:		__db_errx(dbenv,
+msgerr:		__db_errx(env,
 		    "Recovery function for LSN %lu %lu failed on %s pass",
 		    (u_long)lsn.file, (u_long)lsn.offset, pass);
 	}
 
-err:	if (logc != NULL && (t_ret = __log_c_close(logc)) != 0 && ret == 0)
+err:	if (logc != NULL && (t_ret = __logc_close(logc)) != 0 && ret == 0)
 		ret = t_ret;
 
 	if (txninfo != NULL)
-		__db_txnlist_end(dbenv, txninfo);
+		__db_txnlist_end(env, txninfo);
 
 	dbenv->tx_timestamp = 0;
 
-	F_CLR(dbenv->lg_handle, DBLOG_RECOVER);
+	F_CLR(env->lg_handle, DBLOG_RECOVER);
 	F_CLR(region, TXN_IN_RECOVERY);
 
 	return (ret);
 }
 
+#ifndef HAVE_BREW
 /*
  * Figure out how many logfiles we have processed.  If we are moving
  * forward (is_forward != 0), then we're computing current - low.  If
@@ -658,6 +684,7 @@ __lsn_diff(low, high, current, max, is_forward)
 	}
 	return (nf);
 }
+#endif
 
 /*
  * __log_backup --
@@ -668,18 +695,18 @@ __lsn_diff(low, high, current, max, is_forward)
  * Also used in the verify phase to walk back via checkpoints.
  *
  * Find the latest checkpoint whose ckp_lsn is less than the max lsn.
- * PUBLIC: int    __log_backup __P((DB_ENV *, DB_LOGC *, DB_LSN *,
+ * PUBLIC: int    __log_backup __P((ENV *, DB_LOGC *, DB_LSN *,
  * PUBLIC:    DB_LSN *, u_int32_t));
  */
 int
-__log_backup(dbenv, logc, max_lsn, start_lsn, cmp)
-	DB_ENV *dbenv;
+__log_backup(env, logc, max_lsn, start_lsn, cmp)
+	ENV *env;
 	DB_LOGC *logc;
 	DB_LSN *max_lsn, *start_lsn;
 	u_int32_t cmp;
 {
-	DB_LSN cmp_lsn, lsn;
 	DBT data;
+	DB_LSN cmp_lsn, lsn;
 	__txn_ckp_args *ckp_args;
 	int lcmp, ret;
 
@@ -689,14 +716,15 @@ __log_backup(dbenv, logc, max_lsn, start_lsn, cmp)
 	if (cmp != CKPLSN_CMP && cmp != LASTCKP_CMP)
 		return (EINVAL);
 
-	if ((ret = __txn_getckp(dbenv, &lsn)) != 0)
+	if ((ret = __txn_getckp(env, &lsn)) != 0)
 		goto err;
 	/*
 	 * Cmp tells us whether to check the ckp_lsn or the last_ckp
 	 * fields in the checkpoint record.
 	 */
-	while ((ret = __log_c_get(logc, &lsn, &data, DB_SET)) == 0) {
-		if ((ret = __txn_ckp_read(dbenv, data.data, &ckp_args)) != 0)
+	while ((ret = __logc_get(logc, &lsn, &data, DB_SET)) == 0) {
+		if ((ret = __txn_ckp_read(
+		    env, data.data, &ckp_args)) != 0)
 			return (ret);
 		if (cmp == CKPLSN_CMP) {
 			/*
@@ -729,12 +757,12 @@ __log_backup(dbenv, logc, max_lsn, start_lsn, cmp)
 			ret = DB_NOTFOUND;
 			break;
 		}
-		__os_free(dbenv, ckp_args);
+		__os_free(env, ckp_args);
 		ckp_args = NULL;
 	}
 
 	if (ckp_args != NULL)
-		__os_free(dbenv, ckp_args);
+		__os_free(env, ckp_args);
 	/*
 	 * For CKPLSN_CMP if we walked back through all the checkpoints,
 	 * set the cursor on the first log record.  For LASTCKP_CMP
@@ -742,7 +770,7 @@ __log_backup(dbenv, logc, max_lsn, start_lsn, cmp)
 	 */
 err:	if (IS_ZERO_LSN(*start_lsn) && cmp == CKPLSN_CMP &&
 	    (ret == 0 || ret == DB_NOTFOUND))
-		ret = __log_c_get(logc, start_lsn, &data, DB_FIRST);
+		ret = __logc_get(logc, start_lsn, &data, DB_FIRST);
 	return (ret);
 }
 
@@ -754,15 +782,15 @@ err:	if (IS_ZERO_LSN(*start_lsn) && cmp == CKPLSN_CMP &&
  * whose checkpoint LSN is greater than the first LSN we process.
  */
 static int
-__log_earliest(dbenv, logc, lowtime, lowlsn)
-	DB_ENV *dbenv;
+__log_earliest(env, logc, lowtime, lowlsn)
+	ENV *env;
 	DB_LOGC *logc;
 	int32_t *lowtime;
 	DB_LSN *lowlsn;
 {
+	__txn_ckp_args *ckpargs;
 	DB_LSN first_lsn, lsn;
 	DBT data;
-	__txn_ckp_args *ckpargs;
 	u_int32_t rectype;
 	int cmp, ret;
 
@@ -772,17 +800,18 @@ __log_earliest(dbenv, logc, lowtime, lowlsn)
 	 * Read forward through the log looking for the first checkpoint
 	 * record whose ckp_lsn is greater than first_lsn.
 	 */
-	for (ret = __log_c_get(logc, &first_lsn, &data, DB_FIRST);
-	    ret == 0; ret = __log_c_get(logc, &lsn, &data, DB_NEXT)) {
-		memcpy(&rectype, data.data, sizeof(rectype));
+	for (ret = __logc_get(logc, &first_lsn, &data, DB_FIRST);
+	    ret == 0; ret = __logc_get(logc, &lsn, &data, DB_NEXT)) {
+		LOGCOPY_32(env, &rectype, data.data);
 		if (rectype != DB___txn_ckp)
 			continue;
-		if ((ret = __txn_ckp_read(dbenv, data.data, &ckpargs)) == 0) {
+		if ((ret =
+		    __txn_ckp_read(env, data.data, &ckpargs)) == 0) {
 			cmp = LOG_COMPARE(&ckpargs->ckp_lsn, &first_lsn);
 			*lowlsn = ckpargs->ckp_lsn;
 			*lowtime = ckpargs->timestamp;
 
-			__os_free(dbenv, ckpargs);
+			__os_free(env, ckpargs);
 			if (cmp >= 0)
 				break;
 		}
@@ -804,55 +833,62 @@ __log_earliest(dbenv, logc, lowtime, lowlsn)
  * If we are not doing feedback processing (i.e., we are doing txn_recover
  * processing and in_recovery is zero), then last_lsn can be NULL.
  *
- * PUBLIC: int __env_openfiles __P((DB_ENV *, DB_LOGC *,
- * PUBLIC:     void *, DBT *, DB_LSN *, DB_LSN *, double, int));
+ * PUBLIC: int __env_openfiles __P((ENV *,
+ * PUBLIC:     DB_LOGC *, void *, DBT *, DB_LSN *, DB_LSN *, double, int));
  */
 int
-__env_openfiles(dbenv, logc, txninfo,
+__env_openfiles(env, logc, txninfo,
     data, open_lsn, last_lsn, nfiles, in_recovery)
-	DB_ENV *dbenv;
+	ENV *env;
 	DB_LOGC *logc;
 	void *txninfo;
 	DBT *data;
 	DB_LSN *open_lsn, *last_lsn;
-	int in_recovery;
 	double nfiles;
+	int in_recovery;
 {
+	DB_ENV *dbenv;
 	DB_LSN lsn, tlsn;
 	u_int32_t log_size;
 	int progress, ret;
 
+	dbenv = env->dbenv;
+
 	/*
 	 * XXX
 	 * Get the log size.  No locking required because we're single-threaded
 	 * during recovery.
 	 */
-	log_size = ((LOG *)dbenv->lg_handle->reginfo.primary)->log_size;
+	log_size = ((LOG *)env->lg_handle->reginfo.primary)->log_size;
 
 	lsn = *open_lsn;
 	for (;;) {
+#ifdef HAVE_BREW
+		COMPQUIET(nfiles, (double)0.001);
+		COMPQUIET(progress, 0);
+#else
 		if (in_recovery && dbenv->db_feedback != NULL) {
-			DB_ASSERT(dbenv, last_lsn != NULL);
+			DB_ASSERT(env, last_lsn != NULL);
 			progress = (int)(33 * (__lsn_diff(open_lsn,
 			   last_lsn, &lsn, log_size, 1) / nfiles));
 			dbenv->db_feedback(dbenv, DB_RECOVER, progress);
 		}
+#endif
 		tlsn = lsn;
-		ret = __db_dispatch(dbenv,
-		    dbenv->recover_dtab, dbenv->recover_dtab_size, data, &tlsn,
+		ret = __db_dispatch(env, &env->recover_dtab, data, &tlsn,
 		    in_recovery ? DB_TXN_OPENFILES : DB_TXN_POPENFILES,
 		    txninfo);
 		if (ret != 0 && ret != DB_TXN_CKP) {
-			__db_errx(dbenv,
+			__db_errx(env,
 			    "Recovery function for LSN %lu %lu failed",
 			    (u_long)lsn.file, (u_long)lsn.offset);
 			break;
 		}
-		if ((ret = __log_c_get(logc, &lsn, data, DB_NEXT)) != 0) {
+		if ((ret = __logc_get(logc, &lsn, data, DB_NEXT)) != 0) {
 			if (ret == DB_NOTFOUND) {
 				if (last_lsn != NULL &&
 				   LOG_COMPARE(&lsn, last_lsn) != 0)
-					ret = __db_log_corrupt(dbenv, &lsn);
+					ret = __db_log_corrupt(env, &lsn);
 				else
 					ret = 0;
 			}
@@ -864,11 +900,11 @@ __env_openfiles(dbenv, logc, txninfo,
 }
 
 static int
-__db_log_corrupt(dbenv, lsnp)
-	DB_ENV *dbenv;
+__db_log_corrupt(env, lsnp)
+	ENV *env;
 	DB_LSN *lsnp;
 {
-	__db_errx(dbenv, "Log file corrupt at LSN: [%lu][%lu]",
+	__db_errx(env, "Log file corrupt at LSN: [%lu][%lu]",
 	     (u_long)lsnp->file, (u_long)lsnp->offset);
 	return (EINVAL);
 }
@@ -876,11 +912,11 @@ __db_log_corrupt(dbenv, lsnp)
 /*
  * __env_init_rec --
  *
- * PUBLIC: int __env_init_rec __P((DB_ENV *, u_int32_t));
+ * PUBLIC: int __env_init_rec __P((ENV *, u_int32_t));
  */
 int
-__env_init_rec(dbenv, version)
-	DB_ENV *dbenv;
+__env_init_rec(env, version)
+	ENV *env;
 	u_int32_t version;
 {
 	int ret;
@@ -890,128 +926,190 @@ __env_init_rec(dbenv, version)
 	 * functions.  Then we overwrite only specific entries based on
 	 * each previous version we support.
 	 */
-	if ((ret = __env_init_rec_45(dbenv)) != 0)
-		return (ret);
-	ret = 0;
-	switch (version) {
+	if ((ret = __bam_init_recover(env, &env->recover_dtab)) != 0)
+		goto err;
+	if ((ret = __crdel_init_recover(env, &env->recover_dtab)) != 0)
+		goto err;
+	if ((ret = __db_init_recover(env, &env->recover_dtab)) != 0)
+		goto err;
+	if ((ret = __dbreg_init_recover(env, &env->recover_dtab)) != 0)
+		goto err;
+	if ((ret = __fop_init_recover(env, &env->recover_dtab)) != 0)
+		goto err;
+	if ((ret = __ham_init_recover(env, &env->recover_dtab)) != 0)
+		goto err;
+	if ((ret = __qam_init_recover(env, &env->recover_dtab)) != 0)
+		goto err;
+	if ((ret = __txn_init_recover(env, &env->recover_dtab)) != 0)
+		goto err;
+
 	/*
-	 * There are no log record/recovery differences between
-	 * 4.4 and 4.5.  The log version changed due to checksum.
+	 * After installing all the current recovery routines, we want to
+	 * override them with older versions if we are reading a down rev
+	 * log (from a downrev replication master).  If a log record is
+	 * changed then we must use the previous version for all older
+	 * logs.  If a record is changed in multiple revisions then the
+	 * oldest revision that applies must be used.  Therefore we override
+	 * the recovery functions in reverse log version order.
 	 */
-	case DB_LOGVERSION_45:
-	case DB_LOGVERSION_44:
-		break;
-	case DB_LOGVERSION_43:
-		ret = __env_init_rec_43(dbenv);
-		break;
-	case DB_LOGVERSION_42:
-		ret = __env_init_rec_42(dbenv);
-		break;
-	default:
-		__db_errx(dbenv, "Unknown version %lu", (u_long)version);
+	if (version == DB_LOGVERSION)
+		goto done;
+	if ((ret = __env_init_rec_48(env)) != 0)
+		goto err;
+	/*
+	 * Patch 2 added __db_pg_trunc but did not replace any log records
+	 * so we want to override the same functions as in the original release.
+	 */
+	if (version >= DB_LOGVERSION_48)
+		goto done;
+	if ((ret = __env_init_rec_47(env)) != 0)
+		goto err;
+	if (version == DB_LOGVERSION_47)
+		goto done;
+	if ((ret = __env_init_rec_46(env)) != 0)
+		goto err;
+	/*
+	 * There are no log record/recovery differences between 4.4 and 4.5.
+	 * The log version changed due to checksum.  There are no log recovery
+	 * differences between 4.5 and 4.6.  The name of the rep_gen in
+	 * txn_checkpoint changed (to spare, since we don't use it anymore).
+	 */
+	if (version >= DB_LOGVERSION_44)
+		goto done;
+	if ((ret = __env_init_rec_43(env)) != 0)
+		goto err;
+	if (version == DB_LOGVERSION_43)
+		goto done;
+	if (version != DB_LOGVERSION_42) {
+		__db_errx(env, "Unknown version %lu", (u_long)version);
 		ret = EINVAL;
-		break;
+		goto err;
 	}
-	return (ret);
+	ret = __env_init_rec_42(env);
+
+done:
+err:	return (ret);
 }
 
 static int
-__env_init_rec_42(dbenv)
-	DB_ENV *dbenv;
+__env_init_rec_42(env)
+	ENV *env;
 {
 	int ret;
 
-	if ((ret = __db_add_recovery(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size, __db_relink_42_recover,
-	    DB___db_relink_42)) != 0)
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __db_relink_42_recover, DB___db_relink_42)) != 0)
 		goto err;
-	if ((ret = __db_add_recovery(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size, __db_pg_alloc_42_recover,
-	    DB___db_pg_alloc_42)) != 0)
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __db_pg_alloc_42_recover, DB___db_pg_alloc_42)) != 0)
 		goto err;
-	if ((ret = __db_add_recovery(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size, __db_pg_free_42_recover,
-	    DB___db_pg_free_42)) != 0)
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __db_pg_free_42_recover, DB___db_pg_free_42)) != 0)
 		goto err;
-	if ((ret = __db_add_recovery(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size, __db_pg_freedata_42_recover,
-	    DB___db_pg_freedata_42)) != 0)
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __db_pg_freedata_42_recover, DB___db_pg_freedata_42)) != 0)
 		goto err;
-	if ((ret = __db_add_recovery(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size, __ham_metagroup_42_recover,
-	    DB___ham_metagroup_42)) != 0)
+#ifdef HAVE_HASH
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __ham_metagroup_42_recover, DB___ham_metagroup_42)) != 0)
 		goto err;
-	if ((ret = __db_add_recovery(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size, __ham_groupalloc_42_recover,
-	    DB___ham_groupalloc_42)) != 0)
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __ham_groupalloc_42_recover, DB___ham_groupalloc_42)) != 0)
 		goto err;
-	if ((ret = __db_add_recovery(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size, __txn_ckp_42_recover,
-	    DB___txn_ckp_42)) != 0)
-		goto err;
-	if ((ret = __db_add_recovery(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size, __txn_regop_42_recover,
-	    DB___txn_regop_42)) != 0)
+#endif
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __txn_ckp_42_recover, DB___txn_ckp_42)) != 0)
 		goto err;
 err:
 	return (ret);
 }
 
 static int
-__env_init_rec_43(dbenv)
-	DB_ENV *dbenv;
+__env_init_rec_43(env)
+	ENV *env;
 {
 	int ret;
 
-	if ((ret = __db_add_recovery(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size, __bam_relink_43_recover,
-	    DB___bam_relink_43)) != 0)
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __bam_relink_43_recover, DB___bam_relink_43)) != 0)
 		goto err;
 	/*
 	 * We want to use the 4.2-based txn_regop record.
 	 */
-	if ((ret = __db_add_recovery(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size, __txn_regop_42_recover,
-	    DB___txn_regop_42)) != 0)
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __txn_regop_42_recover, DB___txn_regop_42)) != 0)
 		goto err;
 err:
 	return (ret);
 }
 
-/*
- * __env_init_rec_45 --
- *
- */
 static int
-__env_init_rec_45(dbenv)
-	DB_ENV *dbenv;
+__env_init_rec_46(env)
+	ENV *env;
 {
 	int ret;
 
-	if ((ret = __bam_init_recover(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size)) != 0)
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __bam_merge_44_recover, DB___bam_merge_44)) != 0)
 		goto err;
-	if ((ret = __crdel_init_recover(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size)) != 0)
+
+err:	return (ret);
+}
+
+static int
+__env_init_rec_47(env)
+	ENV *env;
+{
+	int ret;
+
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __bam_split_42_recover, DB___bam_split_42)) != 0)
 		goto err;
-	if ((ret = __db_init_recover(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size)) != 0)
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __db_pg_sort_44_recover, DB___db_pg_sort_44)) != 0)
 		goto err;
-	if ((ret = __dbreg_init_recover(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size)) != 0)
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __fop_create_42_recover, DB___fop_create_42)) != 0)
 		goto err;
-	if ((ret = __fop_init_recover(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size)) != 0)
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __fop_write_42_recover, DB___fop_write_42)) != 0)
 		goto err;
-	if ((ret = __ham_init_recover(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size)) != 0)
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __fop_rename_42_recover, DB___fop_rename_42)) != 0)
 		goto err;
-	if ((ret = __qam_init_recover(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size)) != 0)
-		goto err;
-	if ((ret = __txn_init_recover(dbenv, &dbenv->recover_dtab,
-	    &dbenv->recover_dtab_size)) != 0)
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __fop_rename_noundo_46_recover, DB___fop_rename_noundo_46)) != 0)
 		goto err;
+
+err:
+	return (ret);
+}
+
+static int
+__env_init_rec_48(env)
+	ENV *env;
+{
+	int ret;
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __db_pg_sort_44_recover, DB___db_pg_sort_44)) != 0)
+		goto err;
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __db_addrem_42_recover, DB___db_addrem_42)) != 0)
+		goto err;
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __db_big_42_recover, DB___db_big_42)) != 0)
+		goto err;
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __bam_split_48_recover, DB___bam_split_48)) != 0)
+		goto err;
+#ifdef HAVE_HASH
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __ham_insdel_42_recover, DB___ham_insdel_42)) != 0)
+		goto err;
+	if ((ret = __db_add_recovery_int(env, &env->recover_dtab,
+	    __ham_replace_42_recover, DB___ham_replace_42)) != 0)
+		goto err;
+#endif
 err:
 	return (ret);
 }
